# check.py | commands to check answers
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import string
from difflib import get_close_matches

import discord
import discord.ext.commands.view
from discord import app_commands
from discord.ext import commands

from sciolyid.data import (
    database,
    get_aliases,
    format_wiki_url,
    logger,
    possible_words,
    prompts,
)
from sciolyid.data_functions import (
    incorrect_increment,
    item_setup,
    score_increment,
    session_increment,
    streak_increment,
)
from sciolyid.functions import CustomCooldown
from sciolyid.util import better_spellcheck


class Check(commands.Cog):
    def __init__(self, bot):
        self.bot = bot

    # Check command - argument is the guess
    @commands.hybrid_command(
        help="- Checks your answer.", usage="guess", aliases=["guess", "c"]
    )
    @commands.check(CustomCooldown(3.0, bucket=commands.BucketType.user))
    @app_commands.rename(arg="guess")
    @app_commands.describe(arg="your answer")
    async def check(self, ctx: commands.Context, *, arg: str):
        logger.info("command: check")

        current_item = database.hget(f"channel:{ctx.channel.id}", "item").decode(
            "utf-8"
        )
        if current_item == "":  # no image
            await ctx.send("You must ask for a image first!")
            return

        # if there is a image, it checks answer
        arg = arg.lower()
        current_item = current_item.lower()
        logger.info("current_item: " + current_item)
        logger.info("arg: " + arg)

        item_setup(ctx, current_item)
        correct_list = (x.lower() for x in get_aliases(current_item))

        if database.exists(f"race.data:{ctx.channel.id}"):
            logger.info("race in session")
            if database.hget(f"race.data:{ctx.channel.id}", "strict"):
                logger.info("strict spelling")
                correct = arg in correct_list
            else:
                logger.info("spelling leniency")
                correct = better_spellcheck(arg, correct_list, possible_words)
        else:
            logger.info("no race")
            if database.hget(f"session.data:{ctx.author.id}", "strict"):
                logger.info("strict spelling")
                correct = arg in correct_list
            else:
                logger.info("spelling leniency")
                correct = better_spellcheck(arg, correct_list, possible_words)

        if correct:
            logger.info("correct")

            database.hset(f"channel:{ctx.channel.id}", "item", "")
            database.hset(f"channel:{ctx.channel.id}", "answered", "1")

            session_increment(ctx, "correct", 1)
            streak_increment(ctx, 1)
            database.zincrby(
                f"correct.user:{ctx.author.id}",
                1,
                string.capwords(str(current_item)),
            )

            await ctx.send(
                f"Correct! Good job! The image was **{current_item}**."
                if not database.exists(f"race.data:{ctx.channel.id}")
                else f"**{ctx.author.mention}**, you are correct! The image was **{current_item}**."
            )
            url = format_wiki_url(ctx, current_item)
            await ctx.send(url)  # sends wiki page
            score_increment(ctx, 1)
            if database.exists(f"race.data:{ctx.channel.id}"):

                limit = int(database.hget(f"race.data:{ctx.channel.id}", "limit"))
                first = database.zrevrange(f"race.scores:{ctx.channel.id}", 0, 0, True)[
                    0
                ]
                if int(first[1]) >= limit:
                    logger.info("race ending")
                    race = self.bot.get_cog("Race")
                    await race.stop_race(ctx)
                else:
                    logger.info("auto sending next image")
                    group, state, bw = database.hmget(
                        f"race.data:{ctx.channel.id}", ["group", "state", "bw"]
                    )
                    media = self.bot.get_cog("Media")
                    await media.send_pic(
                        ctx,
                        group.decode("utf-8"),
                        state.decode("utf-8"),
                        bw.decode("utf-8"),
                    )

        elif len(prompts.get(current_item, [])) != 0 and better_spellcheck(
            arg, prompts.get(current_item, []), possible_words
        ):
            logger.info("prompt")
            await ctx.send(
                "Close, but not quite what we were looking for. Can you be more specific?"
            )

        else:
            logger.info("incorrect")

            streak_increment(ctx, None)
            session_increment(ctx, "incorrect", 1)
            incorrect_increment(ctx, str(current_item), 1)

            if database.exists(f"race.data:{ctx.channel.id}"):
                await ctx.send("Sorry, that wasn't the right answer.")
            else:
                database.hset(f"channel:{ctx.channel.id}", "item", "")
                database.hset(f"channel:{ctx.channel.id}", "answered", "1")
                await ctx.send("Sorry, the image was actually " + current_item + ".")
                url = format_wiki_url(ctx, current_item)
                await ctx.send(url)

    async def race_autocheck(self, message: discord.Message):
        if not database.exists(f"race.data:{message.channel.id}"):
            return
        if len(get_close_matches(message.content.strip().lower(), possible_words)) != 0:
            logger.info("race autocheck found: checking")
            ctx = commands.Context(
                message=message,
                bot=self.bot,
                prefix="race-autocheck",
                view=discord.ext.commands.view.StringView(""),
            )
            await self.check(ctx, arg=message.content)


async def setup(bot):
    cog = Check(bot)
    bot.add_message_handler(cog.race_autocheck)
    await bot.add_cog(cog)



# git.py | functions for interacting with git
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

import filelock
import git.repo.fun
from git import Repo

import sciolyid.config as config
from sciolyid.data import logger


def _setup_repo(repo_url: str, repo_dir: str) -> Repo:
    lock = filelock.FileLock(repo_dir.strip("/") + ".lock")
    with lock:
        if os.path.exists(repo_dir) and git.repo.fun.is_git_dir(
            os.path.join(repo_dir, ".git")
        ):
            repo = Repo(repo_dir)
            repo.remote("origin").fetch()
            repo.head.reset(working_tree=True)
        else:
            os.makedirs(repo_dir, exist_ok=True)
            new_repo_url = repo_url.split("/")
            if ":" not in new_repo_url[2] and "@" not in new_repo_url[2]:
                new_repo_url[2] = (
                    os.environ[config.options["git_user_env"]]
                    + ":"
                    + os.environ[config.options["git_token_env"]]
                    + "@"
                    + new_repo_url[2]
                )
            repo = Repo.clone_from("/".join(new_repo_url), repo_dir)

            with repo.config_writer() as cw:
                cw.set_value(
                    "user", "email", os.environ[config.options["git_email_env"]]
                )
                cw.set_value("user", "name", os.environ[config.options["git_user_env"]])

    logger.info("done!")
    return repo


verify_repo: Repo = _setup_repo(
    config.options["validation_repo_url"], config.options["validation_local_dir"]
)
image_repo: Repo = _setup_repo(
    config.options["github_image_repo_url"], config.options["download_dir"]
)



# functions/webhooks.py | helper functions for working with Discord webhooks
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import os

import requests
from discord import Color, Embed, SyncWebhook

import sciolyid.config as config
from sciolyid.web.functions.user import (
    AVATAR_URL,
    PROFILE_URL,
    DiscordBotAuth,
    fetch_profile,
)

WEBHOOK_URL = os.getenv(config.options["discord_webhook_env"], None)

if WEBHOOK_URL:
    webhook = SyncWebhook.from_url(WEBHOOK_URL)

    def bot_info() -> dict:
        url = PROFILE_URL.format(id="@me")
        resp = requests.get(url, auth=DiscordBotAuth(), timeout=10)
        json: dict = resp.json()

        output: dict = {}
        output["avatar"] = (
            AVATAR_URL.format(
                id=json["id"],
                avatar=json["avatar"],
                ext=("gif" if json["avatar"].startswith("a_") else "png"),
            )
            if json["avatar"]
            else None
        )
        output["name"] = json["username"]
        return output

    bot = bot_info()
    BOT_AVATAR = bot["avatar"]
    BOT_NAME = bot["name"]


def send(type_of: str, **opt):
    """Wrapper to send embed webhooks.

    type: str - one of "add", "verify", "valid", "error"

    opt:
    - add:
        - num: int/str - number of images added
        - items: List[str] - id items added
        - user_id: str - user id
    - verify:
        - action: str - one of "valid", "invalid", "duplicate"
        - user_id: str - user id
    - valid:
        - added: int - num added
        - rejected: int - num rejected
        - items: List[str] - id items added
        - urls: Tuple/List[str, str] - git diff links, image first, then verification
    - error:
        - message: str - error message
    """
    if not WEBHOOK_URL:
        return

    if type_of not in ("add", "verify", "valid", "error"):
        raise TypeError("Invalid type_of value!")

    if type_of in config.options["discord_webhook_disable"]:
        return

    if type_of == "add":
        username = fetch_profile(opt["user_id"])["username"]
        items = list(set(opt["items"]))
        items[-1] = "and " + items[-1] if len(items) > 1 else items[-1]

        name = "New images added!"
        content = (
            f"<@{opt['user_id']}> ({username}) added {opt['num']} new "
            + f"image{'' if opt['num'] == 1 else 's'} for {(' ' if len(items)== 2 else ', ').join(items)}!"
        )
        color = Color.dark_green()

    elif type_of == "verify":
        username = fetch_profile(opt["user_id"])["username"]
        color_options = {
            "valid": Color.green(),
            "invalid": Color.red(),
            "duplicate": Color.gold(),
        }

        name = "Image verified!"
        content = (
            f"<@{opt['user_id']}> ({username}) marked an image as {opt['action']}."
        )
        color = color_options[opt["action"]]

    elif type_of == "valid":
        items = list(set(opt["items"]))
        if opt["added"] != 0:
            items[-1] = "and " + items[-1] if len(items) > 1 else items[-1]

        name = "Updated images!"
        content = (
            f"**{opt['added']}** new image{'' if opt['added'] == 1 else 's'} {'was' if opt['added'] == 1 else 'were'} added"
            + (
                "."
                if opt["added"] == 0
                else f" for {(' ' if len(items)== 2 else ', ').join(items)}."
            )
            + f" **{opt['rejected']}** {'was' if opt['rejected'] == 1 else 'were'} rejected.\n"
            + f"[Image Repo Commit]({opt['urls'][0]}) **|** [Verification Repo Commit]({opt['urls'][1]})"
        )
        color = Color.blue()

    elif type_of == "error":
        name = "An error occurred!"
        content = opt["message"]
        color = Color.dark_red()

    embed = Embed(type="rich", color=color, title=name, description=content)
    embed.set_footer(
        text=datetime.datetime.now().strftime("%a, %b %d, %Y at %I:%M:%S %p")
    )

    webhook.send(username=f"{BOT_NAME} Web", avatar_url=BOT_AVATAR, embed=embed)



# functions/images.py | functions for working with images
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import csv
import imghdr
import os
from typing import Dict, Optional, Set, Union

import imagehash
import requests
from flask import abort
from PIL import Image

import sciolyid.config as config
from sciolyid.web.config import logger

VALID_MIMETYPES = ("image/jpeg", "image/png")
VALID_IMG_TYPES = ("jpeg", "png")
MAX_FILESIZE = 4000000  # 4 mb


def find_duplicates(image, distance: int = 5, ignore_verify: bool = False) -> list:
    logger.info("find duplicates")
    files: Set[str] = set()
    for url in config.options["hashes_url"]:
        if (
            ignore_verify
            and "/".join(config.options["validation_repo_url"].split("/")[-2:]).split(
                ".git"
            )[0]
            in url
        ):
            continue
        resp = requests.get(url, timeout=10)
        if resp.status_code != 200:
            logger.info(
                f"hashes lookup failed: status {resp.status_code}; url {resp.url}"
            )
            return ["Failed to get hashes file."]
        files = files.union(set(map(lambda x: x.strip(), resp.text.split("\n"))))
    files.discard("")

    if isinstance(image, str):
        with Image.open(image) as img_obj:
            current_hash = imagehash.phash(img_obj)
    else:
        current_hash = imagehash.phash(image)
    matches = []
    r = csv.reader(files)
    for url, image_hash in r:
        if current_hash - imagehash.hex_to_hash(image_hash) <= distance:
            matches.append(url)
    return matches


def generate_id_lookup(ignore_verify: bool = False) -> Optional[Dict[str, str]]:
    logger.info("generate id lookup")
    files: Set[str] = set()
    for url in config.options["ids_url"]:
        if (
            ignore_verify
            and "/".join(config.options["validation_repo_url"].split("/")[-2:]).split(
                ".git"
            )[0]
            in url
        ):
            continue
        resp = requests.get(url, timeout=10)
        if resp.status_code != 200:
            logger.info(f"id lookup failed: status {resp.status_code}; url {resp.url}")
            return None
        files = files.union(set(map(lambda x: x.strip(), resp.text.split("\n"))))
    files.discard("")

    lookup = {}
    r = csv.reader(files)
    for filename, image_id in r:
        lookup[filename] = image_id
    logger.info(f"num lookup ids: {len(lookup)}")
    return lookup


def filename_lookup(start_path: str) -> dict:
    id_lookup = generate_id_lookup()
    if not id_lookup:
        abort(404, "filename lookup failed!")
    result = {}
    stack = []
    stack.append(start_path)
    while stack:
        current = stack.pop()
        for child_filename in os.listdir(current):
            if child_filename.startswith("."):
                continue
            child_path = os.path.join(current, child_filename)
            if os.path.isdir(child_path):
                stack.append(child_path)
                continue
            if imghdr.what(child_path) in VALID_IMG_TYPES:
                image_id = id_lookup.get("./" + os.path.relpath(child_path, start_path))
                if image_id:
                    result[image_id] = child_path
    logger.info(f"found {len(result)} files")
    return result


def verify_image(f, mimetype) -> Union[bool, str]:
    if mimetype not in VALID_MIMETYPES:
        return False

    f.seek(0, 2)
    size = f.tell()
    f.seek(0)
    if size > MAX_FILESIZE:
        return False

    ext = imghdr.what(None, h=f.read())
    if ext not in VALID_IMG_TYPES:
        return False

    try:
        with Image.open(f) as image:
            image.verify()
    except:  # pylint: disable=bare-except
        return False

    return ext



import setuptools


def readme():
    with open("README.md", "r") as f:
        return f.read()


setuptools.setup(
    name="sciolyid",
    version="2.0.1",
    description="Create ID Discord bots for SciOly studying.",
    long_description=readme(),
    long_description_content_type="text/markdown",
    keywords="science_olympiad discord_bot discord studying",
    license="GPLv3+",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    url="https://github.com/tctree333/SciOly-ID",
    author="Tomi Chen",
    author_email="tomichen33@gmail.com",
    packages=setuptools.find_packages(),
    install_requires=[
        "discord.py>=2.1.0, <3.0.0",
        "redis>=3.3.5, <5.0.0",
        "sentry-sdk>=1.1.0, <2.0.0",
        "Pillow>=9.0.1, <10.0.0",
        "wikipedia>=1.4.0, <2.0.0",
        "gitpython>=3.0.6, <4.0.0",
        "hiredis>=1.0.1, <3.0.0",
        "pandas>=1.3.0, <1.5.0",
        "filelock>=3.4.2, <3.7.0",
    ],
    extras_require={
        "web": [
            "Flask>=1.1.2, <2.1.0",
            "Authlib>=0.15.5, <1.0.0",
            "gunicorn>=20.0.4, <21.0.0",
            "ImageHash>=4.0.0, <5.0.0",
            "Celery>=4.4.3, <5.0.0",
            "blinker>=1.4, <1.5",
        ]
    },
    py_modules=["config", "core", "data", "functions", "downloads", "start_bot"],
    python_requires=">=3.8",
)



import csv

import imagehash


def get_hashlist():
    hashes = []
    filenames = []
    with open("hashes.csv", "r") as f:
        reader = csv.reader(f)
        for filename, img_hash in reader:
            hashes.append(img_hash)
            filenames.append(filename)
    return (hashes, filenames)


def find_duplicates(hashes, filenames):
    duplicates = []
    for i, img_hash in enumerate(hashes):
        matches = [filenames[i]]
        img_hash = imagehash.hex_to_hash(img_hash)
        for j, other_hash in enumerate(hashes):
            other_hash = imagehash.hex_to_hash(other_hash)
            if filenames[i] != filenames[j] and img_hash - other_hash <= 5:
                matches.append(filenames[j])
        if len(matches) > 1:
            duplicates.append(matches)
    return duplicates


print(find_duplicates(*get_hashlist()))
